/*
 *  playlist.c
 *  mod_musicindex
 *
 *  $Id: playlist.c,v 1.35 2003/06/27 19:00:29 varenet Exp $
 *
 *  Created by Thibaut VARENE on Thu Mar 20 2003.
 *  Copyright (c) 2003 Regis BOUDIN
 *  Copyright (c) 2003 Thibaut VARENE
 *   
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1, or (at your option)
 *  any later version.
 *
 */

/**
 * @file 
 * Playlist management system.
 *
 * This is where the magic takes place. Here are the functions in charge of
 * analyzing the content of the folders requested by the client, and of
 * generating the structures and lists that will be used by the HTML
 * subsystem to generate the pretty output.
 *
 * @todo Find out if we trash "detect by content".
 * @todo Maybe we could differenciate files by their types (mp3/ogg) to display
 * them differently - Some players do not support all formats we can stream.
 */

#include "playlist.h"
#include "html.h"

/**
 * MP3 content detection doesn't actually work.
 */
#undef DETECT_BY_CONTENT_NOT_BORKEN
#ifdef DETECT_BY_CONTENT_NOT_BORKEN
/* The following pieces of code are taken from xmms, in:
 *	Input/mpg123/common.c
 *	Input/mpg123/mpg123.c
 */
static unsigned long mpg123_convert_to_header(unsigned char * buf)
{
        return (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
}

/**
 * Checks for valid MP3 header.
 *
 * doesn't seem to work :P
 *
 * @param head The file header to check
 *
 * @return FALSE when the header is correct, TRUE otherwise.
 */
static short mpg123_head_check(unsigned long head)
{
        if ((head & 0xffe00000) != 0xffe00000)
                return FALSE;
        if (!((head >> 17) & 3))
                return FALSE;
        if (((head >> 12) & 0xf) == 0xf)
                return FALSE;
        if (!((head >> 12) & 0xf))
                return FALSE;
        if (((head >> 10) & 0x3) == 0x3)
                return FALSE;
        if (((head >> 19) & 1) == 1 && ((head >> 17) & 3) == 3 && ((head >> 16) & 1) == 1)
                return FALSE;
        if ((head & 0xffff0000) == 0xfffe0000)
                return FALSE;

        return TRUE;
}
#endif

/**
 * Checks for valid MP3 filename extension.
 *
 * @param filename The filename to check.
 *
 * @return FALSE when the extension is correct, TRUE otherwise.
 */
static short mpg123_mp3_ext_check(char *filename)
{
	char *ext = strrchr(filename, '.');
	if (ext	&& (!strncasecmp(ext, ".mp2", 4) || !strncasecmp(ext, ".mp3", 4)))
		return FALSE;
	return TRUE;
}

/**
 * Converts an id3tag string in utf8.
 *
 * This function reads the id3tag from an MP3 file (if any) and converts
 * the result in the utf8 standard format
 * 
 * @param tag The tag struct to parse
 * @param frameid The type of tag data searched
 * @param index The id3tag index value
 *
 * @return When possible, an utf8_t string, NULL otherwise.
 */
static id3_utf8_t *utf8_id3tag_findframe(struct id3_tag *tag, char const *frameid,
	unsigned short index)
{
	const struct id3_frame	*frame = NULL;
	const union id3_field	*field = NULL;
	const id3_ucs4_t	*ucs4 = NULL;

	if ((frame = id3_tag_findframe (tag, frameid, index))) {
		field = &frame->fields[1];
		if (id3_field_getnstrings(field)
			&& (ucs4 = id3_field_getstrings(field, index)))
			return(id3_ucs4_utf8duplicate(ucs4));
			/* id3_ucs4_utf8duplicate() allocates memory -> free()! */
	}
	return NULL;
}

/**
 * Fills in the information fields about OGG data.
 *
 * This function reads the ogg tags (using libvorbisfile) from the ogg file
 * and fills in the struct mu_ent fields accordingly. If the file is not an
 * ogg one, the function returns p = head.
 * 
 * @param pool Pool
 * @param head Head
 * @param in ogg file to parse
 * @param conf MusicIndex configuration paramaters struct
 * @param names Names
 * @param r Apache request_rec struct to handle log writings (debugging)
 *
 * @return When possible, struct mu_ent correctly set up
 */
struct mu_ent *make_ogg_entry(pool *pool, struct mu_ent *head,
	FILE *in, mu_config * conf, mu_ent_names *names, request_rec * r)
{
	struct mu_ent		*p = head;
	OggVorbis_File		vf;
	vorbis_comment		*comment = NULL;
	char 			*t;
		
	if (!(ov_open(in, &vf, NULL, 0))) {
		p = new_ent(pool, head);
		p->filetype = FT_OGG;
	
		if ((comment = ov_comment(&vf, -1))) {
			if ((t = vorbis_comment_query(comment, "album", 0)))
				p->album = ap_pstrdup(pool, t);
			if ((t = vorbis_comment_query(comment, "artist", 0)))
				p->artist = ap_pstrdup(pool, t);
			if ((t = vorbis_comment_query(comment, "title", 0)))
				p->title = ap_pstrdup(pool, t);
			if ((t = vorbis_comment_query(comment, "tracknumber", 0)))
				p->track = atoi(t);
			if ((t = vorbis_comment_query(comment, "date", 0)))
				p->date = atoi(t);
		}

		if (conf->options & (MI_QUICKPL | MI_SEARCH)) {
			p->bitrate = p->length = 0;
		}
		else {
			p->bitrate = (long)ov_bitrate(&vf, -1);
			p->length = (long)ov_time_total(&vf, -1);
		}

		ov_clear(&vf);

	}

	return p;
}

/**
 * Fills in the information fields about MP3 data.
 *
 * This function reads the id3 tags (using libid3tag0) from the MP3 file
 * and fills in the struct mu_ent fields accordingly.
 * 
 * @param pool Pool
 * @param head Head
 * @param in MP3 file to parse
 * @param conf MusicIndex configuration paramaters struct
 * @param names Names
 * @param r Apache request_rec struct to handle log writings (debugging)
 *
 * @return When possible, struct mu_ent correctly set up
 *
 * @todo VBR handling (Xing headers decoding)
 * @todo Check for the use or not of the do() while loop.
 * @warning See the XXX comment: the file extension removing - same for ogg btw.
 */
struct mu_ent *make_mp3_entry(pool *pool, struct mu_ent *head,
	FILE *in, mu_config * conf, mu_ent_names *names, request_rec * r) 
{
	struct mu_ent 		*p = head;

	struct id3_file		*id3struct = NULL;
	struct id3_tag		*tag = NULL;
	id3_utf8_t 		*utf8 = NULL;

	struct stat		filestat;
	
	struct mad_stream 	madstream;
	struct mad_frame	madframe;
	unsigned long		madframe_count = 0;
	unsigned char		madinput_buffer[INPUT_BUFFER_SIZE];

#ifdef DETECT_BY_CONTENT_NOT_BORKEN
	unsigned char		tmp[4];

	if (fread(tmp, 1, 4, in) != 4)
		goto exit;

	if (mpg123_head_check(mpg123_convert_to_header(tmp)))
#endif
	if (mpg123_mp3_ext_check(names->filename))
		goto exit;
	
	mad_stream_init(&madstream);
	mad_frame_init(&madframe);
	
	fstat(fileno(in), &filestat);
	
	/* XXX Compte tenu du fait qu'on a sorti cette initialisation
	de la grosse boucle qui prenait elle bien soin de vrifier le
	de facon minimale l'tat du fichier, on n'est moins sur de la
	veracite des infos fournies... enfin bon 
	j'ajouterai galement que le gain est minime (mais sensible), vu que la
	plus grosse partie des I/O est gnre par la lecture des tags...
	il faudra peut-etre se repencher la dessus, a la lumire de
	l'existence prochaine d'un cache ;) */
	p = new_ent(pool, head);
	p->filetype = FT_MP3;
				
	if (conf->options & (MI_QUICKPL | MI_SEARCH)) {
		p->bitrate = p->length = 0;
	}
	else {
		/* TODO: la boucle est peut-etre inutile, a verifier
		 * peut etre necessaire avec les VBR, a voir...
		 */
		do {
			if (madstream.buffer == NULL || madstream.error == MAD_ERROR_BUFLEN) {
				size_t		madread_size, remaining;
				unsigned char	*madread_start;
				
				if (madstream.next_frame != NULL) {
					remaining = madstream.bufend-madstream.next_frame;
					memmove(madinput_buffer, madstream.next_frame, remaining);
					madread_start=madinput_buffer + remaining;
					madread_size = INPUT_BUFFER_SIZE - remaining;
				}
				else {
					madread_size = INPUT_BUFFER_SIZE;
					madread_start = madinput_buffer;
					remaining = 0;
				}
				
				madread_size = fread(madread_start, 1, madread_size, in);
				if (madread_size <= 0) {
					ap_log_printf(r->server, "maderror madread_size <= 0 on %s", names->filename);
					goto exit;
					/* @todo traitement eventuel erreur lecture */
				}
	
				mad_stream_buffer(&madstream, madinput_buffer, madread_size + remaining);
				madstream.error = 0;
			}
	
	
				
			if (mad_frame_decode(&madframe,&madstream)) {
				if (MAD_RECOVERABLE(madstream.error)) {
					/* ap_log_error_old("maderror madstream.error recoverable", r->server); */
					/* @todo traitement eventuel erreur recoverable */
					continue;
				}
				else {
					if (madstream.error == MAD_ERROR_BUFLEN)
						continue;
					else {
						ap_log_error_old("maderror madstream.error unrecoverable", r->server);
						/* @todo traitement eventuel erreur unrecoverable */
						goto exit;
					}
				}
			}
		
			if (madframe_count == 0) {
				p->bitrate = madframe.header.bitrate * 1.024;
				p->length = filestat.st_size / (madframe.header.bitrate / 8);
				break;
			}
			
			madframe_count++;
		} while (1);
	}

	if ((id3struct = id3_file_open(names->filename, ID3_FILE_MODE_READONLY))) {
		if ((tag = id3_file_tag(id3struct)) && (tag->frames)) {
			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_TITLE, 0))) {
				p->title = ap_pstrdup(pool, (char *)utf8);
				free(utf8);
			}

			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_ARTIST, 0))) {
				p->artist = ap_pstrdup(pool, (char *)utf8);
				free(utf8);
			}
			
			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_ALBUM, 0))) {
				p->album = ap_pstrdup(pool, (char *)utf8);
				free(utf8);
			}

			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_YEAR, 0))) {
				p->date = atoi((char *)utf8);
				free(utf8);
			}

			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_TRACK, 0))) {
				p->track = atoi((char *)utf8);
				free(utf8);
			}
		
			if ((utf8 = utf8_id3tag_findframe(tag, "TPOS", 0))) {
				p->posn = atoi((char *)utf8);
				free(utf8);
			}
		
			if ((utf8 = utf8_id3tag_findframe(tag, "TLEN", 0))) {
				if ((atoi((char *)utf8) / 1000) > 0)
					p->length = atoi((char *)utf8) / 1000;
				free(utf8);
			}
		}
		id3_file_close (id3struct);
	}

exit:
	return p;
}

/**
 * Add a file (and the content of the directory if requested) to the chain.
 *
 * This function creates a new entry from a file.
 * If the file is a directory and the recursive option is set, all its content
 * is also added.
 * If the file is an ogg vorbis or an mp3, it is simply added.
 * 
 * @param pool Pool
 * @param r Apache request_rec struct to handle log writings (debugging)
 * @param head Head
 * @param conf MusicIndex configuration paramaters struct
 * @param names Names
 *
 * @return When possible, struct mu_ent correctly set up
 */
struct mu_ent *make_music_entry(pool *pool, request_rec * r,
	struct mu_ent *head, mu_config * conf, mu_ent_names *names)
{
	DIR 			*dir;
	struct DIR_TYPE		*dstruct;
	struct mu_ent		*p = head;
	FILE			*in = NULL;
	unsigned short		i;
	char			*t, *fn, *uri;

	if (!names) {
		names = ap_palloc(r->pool, sizeof(mu_ent_names));
		names->filename = ap_palloc(r->pool, MAX_STRING);
		names->uri = ap_palloc(r->pool, MAX_STRING);
		strcpy(names->filename, r->filename);
		strcpy(names->uri, r->parsed_uri.path);
	}

	t = names->uri + strlen(names->uri)-1;
	while (*t != '/')
		t--;
	t++;
	if (*t == '.')
		return p;

	if ((in = fopen(names->filename, "r"))) {
		if ((ap_is_directory(names->filename)) ) {
			fn = names->filename + strlen(names->filename);
			*fn++ = '/';
			*fn = '\0';
		
			uri = names->uri + strlen(names->uri) - 1;
			if (*uri++ != '/')
				*uri++ = '/';
			*uri = '\0';
				
			if (conf->options & MI_RECURSIVE) {
				conf->options &= (conf->play_recursive);

				if (!(dir = opendir(names->filename)))
					return head;
				while ((dstruct = readdir(dir))) {
					strcpy(fn, dstruct->d_name);
					strcpy(uri, dstruct->d_name);
					p = make_music_entry(pool, r, p, conf, names);
				}
				closedir(dir);
				fclose(in);
				return p;
			}
			else if (!(conf->options & MI_PLAYLIST)){
				p = new_ent(pool, head);
				p->filetype = FT_DIR;
			}
			fclose(in);
		}
		else if ((p = make_ogg_entry(pool, head, in, conf, names, r)) != head);
		/* ATTENTION! ov_open() takes complete possession of the file,
		   and ov_clear() closes it if ov_open() was successful.
		   We should NOT use fclose() in that case. */
		else if ((p = make_mp3_entry(pool, head, in, conf, names, r)) != head)
			fclose(in);
		else
			fclose(in);
	}

	if (p == head)
		return p;
	p->uri = ap_pstrdup(pool, names->uri);
	p->file = p->uri + strlen(r->parsed_uri.path);
	
	if (p->filetype == FT_DIR)
		return p;

	if (!(p->title)) {
#ifdef NO_TITLE_STRIP
		p->title = p->file;
#else
		/* copy the name removing file extension and changing '_' to ' ' */
		p->title = ap_pstrndup(pool, p->file, strlen(p->file) - 4);
		for (i=0; p->title[i] != '\0'; i++)
			if (p->title[i] == '_')
				p->title[i] = ' ';
#endif
		/* We remove the trailing path if any. ap_filename ? */
		if (strrchr(p->title, '/')) {
			p->title = strrchr(p->title, '/');
			(p->title)++;
		}

	}
		
	if (!(conf->options & MI_SEARCH))
		return p;

	/* Here we go if we have an ongoing search request.
	 * We will return only the entries matching the search criteria */
	if (p->file && (ap_strcasestr(p->file, conf->search)))
		return p;
	if (p->artist && (ap_strcasestr(p->artist, conf->search)))
		return p;
	if (p->album && (ap_strcasestr(p->album, conf->search)))
		return p;
	if (p->title && (ap_strcasestr(p->title, conf->search)))
		return p;

	/* TODO : Clear ent structure. Apache doesn't like the usual "free"
	function */
	return head;
}

/**
 * Generates and sends a webpage listing the music files in the directory.
 *
 * This function reads the id3 tags (using libid3tag0) from the MP3 file
 * and fills in the struct mu_ent fields accordingly.
 * 
 * @param r Apache request_rec struct
 * @param conf MusicIndex configuration paramaters struct
 *
 * @return 0
 */
short musicindex_directory(request_rec * r, mu_config * conf)
{
	DIR		*dir;
	struct mu_ent	*head = NULL;

	if (!(dir = ap_popendir(r->pool, r->filename))) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
			"Can't open directory for music index: %s",
			r->filename);
		return HTTP_FORBIDDEN;
	}

	r->content_type = "text/html; charset=UTF-8";
	ap_send_http_header(r);

	if (r->header_only) {
		ap_pclosedir(r->pool, dir);
		return 0;
	}
	ap_hard_timeout("send music list", r);

	head = make_music_entry(r->pool, r, NULL, conf, NULL);

	head = quicksort(head, NULL, conf);

	send_head(r, conf);

	if (!(conf->options & MI_SEARCH))
		send_directories(r, head, conf);
	
	send_tracks(r, head, conf);
	
	send_foot(r, conf);

	ap_pclosedir(r->pool, dir);

	ap_kill_timeout(r);
	return 0;
}

/**
 * Creates and sends a playlist with the content of a directory.
 *
 * This function generates a playlist containing all the files in the current
 * directory, either recursively or not.
 * 
 * @param r Apache request_rec struct
 * @param conf MusicIndex configuration paramaters struct
 *
 * @return 0
 */
short playlist_directory(request_rec * r, mu_config * conf)
{
	DIR		*dir;
	struct mu_ent	*head = NULL;

	if (!(dir = ap_popendir(r->pool, r->filename)))
		return HTTP_FORBIDDEN;

	r->content_type = "audio/mpegurl";

	ap_send_http_header(r);

	conf->options |= MI_PLAYLIST;

	if (r->header_only) {
		ap_pclosedir(r->pool, dir);
		return 0;
	}
	ap_hard_timeout("send playlist", r);

	head = make_music_entry(r->pool, r, head, conf, NULL);

	head = quicksort(head, NULL, conf);

	send_playlist(r, head, conf);

	ap_pclosedir(r->pool, dir);

	ap_kill_timeout(r);
	return 0;
}

/**
 * Generates and sends a playlist with only selected items.
 *
 * This function parses the requests, which is the form :
 * http://...../?file=file1&file=file2&file=file3...
 *
 * Note that in FORM requests, arguments have ' ' replaced by '+', other
 * "special" characters are changed to their URI equivalent. Here is the trick :
 * ap_unescape_url doesn't change the '+' to its URI equivalent. So we first
 * change ' ' to '+' and then call the apache function. Easy, isn't it ?
 * 
 * @param r Apache request_rec struct
 * @param conf MusicIndex configuration paramaters struct
 *
 * @return 0 
 * @bug XXX Pkoi faire des fonctions qui retournent "0"? void C pas aussi bien?
 */
short playlist_selected(request_rec * r, mu_config * conf)
{
	const char		*args = r->args;
	char			*p;
	unsigned short		i;

	char			*fn, *uri;
	struct mu_ent_names	*names;
	struct mu_ent		*head = NULL;

	r->content_type = "audio/mpegurl";

	ap_send_http_header(r);

	if (r->header_only)
		return 0;
	ap_hard_timeout("send playlist", r);

	if ((args == NULL) || (*args == '\0')) {
		ap_kill_timeout(r);
		return 0;
	}

	names = ap_palloc(r->pool, sizeof(mu_ent_names));
	names->filename = ap_palloc(r->pool, MAX_STRING);
	names->uri = ap_palloc(r->pool, MAX_STRING);
	strcpy(names->filename, r->filename);
	strcpy(names->uri, r->parsed_uri.path);

	fn = names->filename + strlen(names->filename);
	*fn++ = '/';
	*fn = '\0';

	uri = names->uri + strlen(names->uri) - 1;
	if (*uri++ != '/')
		*uri++ = '/';
	*uri = '\0';

	while (args[0]){
		p = ap_getword(r->pool, &args, '&');
		if (!strncmp(p, "file=", 5))
		{
			p += 5;
			strcpy(fn, p);
			for (i=0;fn[i]; i++) {
				if (fn[i] == '+') {
					fn[i] = ' ';
				}
			}
			ap_unescape_url(fn);
			strcpy(uri, fn);
			head = make_music_entry(r->pool, r, head, conf, names);
		}
	}

	head = quicksort(head, NULL, conf);
	send_playlist(r, head, conf);

	ap_kill_timeout(r);
	return 0;
}
